"""Base class to manage comms"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.


import comm.base_comm
import traitlets
import traitlets.config


class CommManager(traitlets.config.LoggingConfigurable, comm.base_comm.CommManager):

    kernel = traitlets.Instance("ipykernel.kernelbase.Kernel")
    comms = traitlets.Dict()
    targets = traitlets.Dict()

    def __init__(self, **kwargs):
        # CommManager doesn't take arguments, so we explicitly forward arguments
        traitlets.config.LoggingConfigurable.__init__(self, **kwargs)
        comm.base_comm.CommManager.__init__(self)



"""Pickle related utilities. Perhaps this should be called 'can'."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
import typing
import warnings

warnings.warn(
    "ipykernel.pickleutil is deprecated. It has moved to ipyparallel.",
    DeprecationWarning,
    stacklevel=2,
)

import copy
import pickle
import sys
from types import FunctionType

# This registers a hook when it's imported
from ipyparallel.serialize import codeutil  # noqa F401
from traitlets.log import get_logger
from traitlets.utils.importstring import import_item

buffer = memoryview
class_type = type

PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL


def _get_cell_type(a=None):
    """the type of a closure cell doesn't seem to be importable,
    so just create one
    """

    def inner():
        return a

    return type(inner.__closure__[0])  # type:ignore[index]


cell_type = _get_cell_type()

# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------


def interactive(f):
    """decorator for making functions appear as interactively defined.
    This results in the function being linked to the user_ns as globals()
    instead of the module globals().
    """

    # build new FunctionType, so it can have the right globals
    # interactive functions never have closures, that's kind of the point
    if isinstance(f, FunctionType):
        mainmod = __import__("__main__")
        f = FunctionType(
            f.__code__,
            mainmod.__dict__,
            f.__name__,
            f.__defaults__,
        )
    # associate with __main__ for uncanning
    f.__module__ = "__main__"
    return f


def use_dill():
    """use dill to expand serialization support

    adds support for object methods and closures to serialization.
    """
    # import dill causes most of the magic
    import dill

    # dill doesn't work with cPickle,
    # tell the two relevant modules to use plain pickle

    global pickle
    pickle = dill

    try:
        from ipykernel import serialize
    except ImportError:
        pass
    else:
        serialize.pickle = dill  # type:ignore[attr-defined]

    # disable special function handling, let dill take care of it
    can_map.pop(FunctionType, None)


def use_cloudpickle():
    """use cloudpickle to expand serialization support

    adds support for object methods and closures to serialization.
    """
    import cloudpickle

    global pickle
    pickle = cloudpickle

    try:
        from ipykernel import serialize
    except ImportError:
        pass
    else:
        serialize.pickle = cloudpickle  # type:ignore[attr-defined]

    # disable special function handling, let cloudpickle take care of it
    can_map.pop(FunctionType, None)


# -------------------------------------------------------------------------------
# Classes
# -------------------------------------------------------------------------------


class CannedObject:
    def __init__(self, obj, keys=None, hook=None):
        """can an object for safe pickling

        Parameters
        ----------
        obj
            The object to be canned
        keys : list (optional)
            list of attribute names that will be explicitly canned / uncanned
        hook : callable (optional)
            An optional extra callable,
            which can do additional processing of the uncanned object.

        Notes
        -----
        large data may be offloaded into the buffers list,
        used for zero-copy transfers.
        """
        self.keys = keys or []
        self.obj = copy.copy(obj)
        self.hook = can(hook)
        for key in keys:
            setattr(self.obj, key, can(getattr(obj, key)))

        self.buffers = []

    def get_object(self, g=None):
        if g is None:
            g = {}
        obj = self.obj
        for key in self.keys:
            setattr(obj, key, uncan(getattr(obj, key), g))

        if self.hook:
            self.hook = uncan(self.hook, g)
            self.hook(obj, g)
        return self.obj


class Reference(CannedObject):
    """object for wrapping a remote reference by name."""

    def __init__(self, name):
        if not isinstance(name, str):
            raise TypeError("illegal name: %r" % name)
        self.name = name
        self.buffers = []

    def __repr__(self):
        return "<Reference: %r>" % self.name

    def get_object(self, g=None):
        if g is None:
            g = {}

        return eval(self.name, g)


class CannedCell(CannedObject):
    """Can a closure cell"""

    def __init__(self, cell):
        self.cell_contents = can(cell.cell_contents)

    def get_object(self, g=None):
        cell_contents = uncan(self.cell_contents, g)

        def inner():
            return cell_contents

        return inner.__closure__[0]  # type:ignore[index]


class CannedFunction(CannedObject):
    def __init__(self, f):
        self._check_type(f)
        self.code = f.__code__
        self.defaults: typing.Optional[typing.List[typing.Any]]
        if f.__defaults__:
            self.defaults = [can(fd) for fd in f.__defaults__]
        else:
            self.defaults = None

        self.closure: typing.Any
        closure = f.__closure__
        if closure:
            self.closure = tuple(can(cell) for cell in closure)
        else:
            self.closure = None

        self.module = f.__module__ or "__main__"
        self.__name__ = f.__name__
        self.buffers = []

    def _check_type(self, obj):
        assert isinstance(obj, FunctionType), "Not a function type"

    def get_object(self, g=None):
        # try to load function back into its module:
        if not self.module.startswith("__"):
            __import__(self.module)
            g = sys.modules[self.module].__dict__

        if g is None:
            g = {}
        if self.defaults:
            defaults = tuple(uncan(cfd, g) for cfd in self.defaults)
        else:
            defaults = None
        if self.closure:
            closure = tuple(uncan(cell, g) for cell in self.closure)
        else:
            closure = None
        newFunc = FunctionType(self.code, g, self.__name__, defaults, closure)
        return newFunc


class CannedClass(CannedObject):
    def __init__(self, cls):
        self._check_type(cls)
        self.name = cls.__name__
        self.old_style = not isinstance(cls, type)
        self._canned_dict = {}
        for k, v in cls.__dict__.items():
            if k not in ("__weakref__", "__dict__"):
                self._canned_dict[k] = can(v)
        if self.old_style:
            mro = []
        else:
            mro = cls.mro()

        self.parents = [can(c) for c in mro[1:]]
        self.buffers = []

    def _check_type(self, obj):
        assert isinstance(obj, class_type), "Not a class type"

    def get_object(self, g=None):
        parents = tuple(uncan(p, g) for p in self.parents)
        return type(self.name, parents, uncan_dict(self._canned_dict, g=g))


class CannedArray(CannedObject):
    def __init__(self, obj):
        from numpy import ascontiguousarray

        self.shape = obj.shape
        self.dtype = obj.dtype.descr if obj.dtype.fields else obj.dtype.str
        self.pickled = False
        if sum(obj.shape) == 0:
            self.pickled = True
        elif obj.dtype == "O":
            # can't handle object dtype with buffer approach
            self.pickled = True
        elif obj.dtype.fields and any(dt == "O" for dt, sz in obj.dtype.fields.values()):
            self.pickled = True
        if self.pickled:
            # just pickle it
            self.buffers = [pickle.dumps(obj, PICKLE_PROTOCOL)]
        else:
            # ensure contiguous
            obj = ascontiguousarray(obj, dtype=None)
            self.buffers = [buffer(obj)]

    def get_object(self, g=None):
        from numpy import frombuffer

        data = self.buffers[0]
        if self.pickled:
            # we just pickled it
            return pickle.loads(data)
        else:
            return frombuffer(data, dtype=self.dtype).reshape(self.shape)


class CannedBytes(CannedObject):
    @staticmethod
    def wrap(buf: typing.Union[memoryview, bytes, typing.SupportsBytes]) -> bytes:
        """Cast a buffer or memoryview object to bytes"""
        if isinstance(buf, memoryview):
            return buf.tobytes()
        if not isinstance(buf, bytes):
            return bytes(buf)
        return buf

    def __init__(self, obj):
        self.buffers = [obj]

    def get_object(self, g=None):
        data = self.buffers[0]
        return self.wrap(data)


class CannedBuffer(CannedBytes):
    wrap = buffer  # type:ignore[assignment]


class CannedMemoryView(CannedBytes):
    wrap = memoryview  # type:ignore[assignment]


# -------------------------------------------------------------------------------
# Functions
# -------------------------------------------------------------------------------


def _import_mapping(mapping, original=None):
    """import any string-keys in a type mapping"""
    log = get_logger()
    log.debug("Importing canning map")
    for key, _ in list(mapping.items()):
        if isinstance(key, str):
            try:
                cls = import_item(key)
            except Exception:
                if original and key not in original:
                    # only message on user-added classes
                    log.error("canning class not importable: %r", key, exc_info=True)
                mapping.pop(key)
            else:
                mapping[cls] = mapping.pop(key)


def istype(obj, check):
    """like isinstance(obj, check), but strict

    This won't catch subclasses.
    """
    if isinstance(check, tuple):
        for cls in check:
            if type(obj) is cls:
                return True
        return False
    else:
        return type(obj) is check


def can(obj):
    """prepare an object for pickling"""

    import_needed = False

    for cls, canner in can_map.items():
        if isinstance(cls, str):
            import_needed = True
            break
        elif istype(obj, cls):
            return canner(obj)

    if import_needed:
        # perform can_map imports, then try again
        # this will usually only happen once
        _import_mapping(can_map, _original_can_map)
        return can(obj)

    return obj


def can_class(obj):
    if isinstance(obj, class_type) and obj.__module__ == "__main__":
        return CannedClass(obj)
    else:
        return obj


def can_dict(obj):
    """can the *values* of a dict"""
    if istype(obj, dict):
        newobj = {}
        for k, v in obj.items():
            newobj[k] = can(v)
        return newobj
    else:
        return obj


sequence_types = (list, tuple, set)


def can_sequence(obj):
    """can the elements of a sequence"""
    if istype(obj, sequence_types):
        t = type(obj)
        return t([can(i) for i in obj])
    else:
        return obj


def uncan(obj, g=None):
    """invert canning"""

    import_needed = False
    for cls, uncanner in uncan_map.items():
        if isinstance(cls, str):
            import_needed = True
            break
        elif isinstance(obj, cls):
            return uncanner(obj, g)

    if import_needed:
        # perform uncan_map imports, then try again
        # this will usually only happen once
        _import_mapping(uncan_map, _original_uncan_map)
        return uncan(obj, g)

    return obj


def uncan_dict(obj, g=None):
    if istype(obj, dict):
        newobj = {}
        for k, v in obj.items():
            newobj[k] = uncan(v, g)
        return newobj
    else:
        return obj


def uncan_sequence(obj, g=None):
    if istype(obj, sequence_types):
        t = type(obj)
        return t([uncan(i, g) for i in obj])
    else:
        return obj


# -------------------------------------------------------------------------------
# API dictionaries
# -------------------------------------------------------------------------------

# These dicts can be extended for custom serialization of new objects

can_map = {
    "numpy.ndarray": CannedArray,
    FunctionType: CannedFunction,
    bytes: CannedBytes,
    memoryview: CannedMemoryView,
    cell_type: CannedCell,
    class_type: can_class,
}
if buffer is not memoryview:
    can_map[buffer] = CannedBuffer

uncan_map: typing.Dict[type, typing.Any] = {
    CannedObject: lambda obj, g: obj.get_object(g),
    dict: uncan_dict,
}

# for use in _import_mapping:
_original_can_map = can_map.copy()
_original_uncan_map = uncan_map.copy()



import os
import sys
import warnings
from unittest import mock

import pytest

from ipykernel.parentpoller import ParentPollerUnix, ParentPollerWindows


@pytest.mark.skipif(os.name == "nt", reason="only works on posix")
def test_parent_poller_unix():
    poller = ParentPollerUnix()
    with mock.patch("os.getppid", lambda: 1):

        def exit_mock(*args):
            sys.exit(1)

        with mock.patch("os._exit", exit_mock), pytest.raises(SystemExit):
            poller.run()

    def mock_getppid():
        raise ValueError("hi")

    with mock.patch("os.getppid", mock_getppid), pytest.raises(ValueError):
        poller.run()


@pytest.mark.skipif(os.name != "nt", reason="only works on windows")
def test_parent_poller_windows():
    poller = ParentPollerWindows(interrupt_handle=1)

    def mock_wait(*args, **kwargs):
        return -1

    with mock.patch("ctypes.windll.kernel32.WaitForMultipleObjects", mock_wait):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            poller.run()



import builtins
import logging
import signal
import threading
import traceback
import warnings

import trio


class TrioRunner:
    def __init__(self):
        self._cell_cancel_scope = None
        self._trio_token = None

    def initialize(self, kernel, io_loop):
        kernel.shell.set_trio_runner(self)
        kernel.shell.run_line_magic("autoawait", "trio")
        kernel.shell.magics_manager.magics["line"]["autoawait"] = lambda _: warnings.warn(
            "Autoawait isn't allowed in Trio background loop mode."
        )
        self._interrupted = False
        bg_thread = threading.Thread(target=io_loop.start, daemon=True, name="TornadoBackground")
        bg_thread.start()

    def interrupt(self, signum, frame):
        if self._cell_cancel_scope:
            self._cell_cancel_scope.cancel()
        else:
            raise Exception("Kernel interrupted but no cell is running")

    def run(self):
        old_sig = signal.signal(signal.SIGINT, self.interrupt)

        def log_nursery_exc(exc):
            exc = "\n".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
            logging.error("An exception occurred in a global nursery task.\n%s", exc)

        async def trio_main():
            self._trio_token = trio.lowlevel.current_trio_token()
            async with trio.open_nursery() as nursery:
                # TODO This hack prevents the nursery from cancelling all child
                # tasks when an uncaught exception occurs, but it's ugly.
                nursery._add_exc = log_nursery_exc
                builtins.GLOBAL_NURSERY = nursery  # type:ignore[attr-defined]
                await trio.sleep_forever()

        trio.run(trio_main)
        signal.signal(signal.SIGINT, old_sig)

    def __call__(self, async_fn):
        async def loc(coro):
            self._cell_cancel_scope = trio.CancelScope()
            with self._cell_cancel_scope:
                return await coro
            self._cell_cancel_scope = None

        return trio.from_thread.run(loc, async_fn, trio_token=self._trio_token)



import os
import threading
import time
from unittest.mock import patch

import pytest

from ipykernel.kernelapp import IPKernelApp

from .conftest import MockKernel

try:
    import trio
except ImportError:
    trio = None


@pytest.mark.skipif(os.name == "nt", reason="requires ipc")
def test_init_ipc_socket():
    app = IPKernelApp(transport="ipc")
    app.init_sockets()
    app.cleanup_connection_file()
    app.close()


def test_blackhole():
    app = IPKernelApp()
    app.no_stderr = True
    app.no_stdout = True
    app.init_blackhole()


def test_start_app():
    app = IPKernelApp()
    app.kernel = MockKernel()

    def trigger_stop():
        time.sleep(1)
        app.io_loop.add_callback(app.io_loop.stop)

    thread = threading.Thread(target=trigger_stop)
    thread.start()
    app.init_sockets()
    app.start()
    app.cleanup_connection_file()
    app.kernel.destroy()
    app.close()


@pytest.mark.skipif(trio is None, reason="requires trio")
def test_trio_loop():
    app = IPKernelApp(trio_loop=True)
    app.kernel = MockKernel()
    app.init_sockets()
    with patch("ipykernel.trio_runner.TrioRunner.run", lambda _: None):
        app.start()
    app.cleanup_connection_file()
    app.io_loop.add_callback(app.io_loop.stop)
    app.kernel.destroy()
    app.close()



"""test the IPython Kernel"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import ast
import os.path
import platform
import signal
import subprocess
import sys
import time
from subprocess import Popen
from tempfile import TemporaryDirectory

import IPython
import psutil
import pytest
from flaky import flaky
from IPython.paths import locate_profile

from .utils import (
    TIMEOUT,
    assemble_output,
    execute,
    flush_channels,
    get_reply,
    kernel,
    new_kernel,
    wait_for_idle,
)


def _check_master(kc, expected=True, stream="stdout"):
    execute(kc=kc, code="import sys")
    flush_channels(kc)
    msg_id, content = execute(kc=kc, code="print(sys.%s._is_master_process())" % stream)
    stdout, stderr = assemble_output(kc.get_iopub_msg)
    assert stdout.strip() == repr(expected)


def _check_status(content):
    """If status=error, show the traceback"""
    if content["status"] == "error":
        assert False, "".join(["\n"] + content["traceback"])


# printing tests


def test_simple_print():
    """simple print statement in kernel"""
    with kernel() as kc:
        msg_id, content = execute(kc=kc, code="print('hi')")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == "hi\n"
        assert stderr == ""
        _check_master(kc, expected=True)


@pytest.mark.skip(reason="Currently don't capture during test as pytest does its own capturing")
def test_capture_fd():
    """simple print statement in kernel"""
    with kernel() as kc:
        iopub = kc.iopub_channel
        msg_id, content = execute(kc=kc, code="import os; os.system('echo capsys')")
        stdout, stderr = assemble_output(iopub)
        assert stdout == "capsys\n"
        assert stderr == ""
        _check_master(kc, expected=True)


@pytest.mark.skip(reason="Currently don't capture during test as pytest does its own capturing")
def test_subprocess_peek_at_stream_fileno():
    with kernel() as kc:
        iopub = kc.iopub_channel
        msg_id, content = execute(
            kc=kc,
            code="import subprocess, sys; subprocess.run(['python', '-c', 'import os; os.system(\"echo CAP1\"); print(\"CAP2\")'], stderr=sys.stderr)",
        )
        stdout, stderr = assemble_output(iopub)
        assert stdout == "CAP1\nCAP2\n"
        assert stderr == ""
        _check_master(kc, expected=True)


def test_sys_path():
    """test that sys.path doesn't get messed up by default"""
    with kernel() as kc:
        msg_id, content = execute(kc=kc, code="import sys; print(repr(sys.path))")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
    # for error-output on failure
    sys.stderr.write(stderr)

    sys_path = ast.literal_eval(stdout.strip())
    assert "" in sys_path


def test_sys_path_profile_dir():
    """test that sys.path doesn't get messed up when `--profile-dir` is specified"""

    with new_kernel(["--profile-dir", locate_profile("default")]) as kc:
        msg_id, content = execute(kc=kc, code="import sys; print(repr(sys.path))")
        stdout, stderr = assemble_output(kc.get_iopub_msg)
    # for error-output on failure
    sys.stderr.write(stderr)

    sys_path = ast.literal_eval(stdout.strip())
    assert "" in sys_path


@flaky(max_runs=3)
@pytest.mark.skipif(
    sys.platform == "win32" or (sys.platform == "darwin" and sys.version_info >= (3, 8)),
    reason="subprocess prints fail on Windows and MacOS Python 3.8+",
)
def test_subprocess_print():
    """printing from forked mp.Process"""
    with new_kernel() as kc:

        _check_master(kc, expected=True)
        flush_channels(kc)
        np = 5
        code = "\n".join(
            [
                "import time",
                "import multiprocessing as mp",
                "pool = [mp.Process(target=print, args=('hello', i,)) for i in range(%i)]" % np,
                "for p in pool: p.start()",
                "for p in pool: p.join()",
                "time.sleep(0.5),",
            ]
        )

        msg_id, content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout.count("hello") == np, stdout
        for n in range(np):
            assert stdout.count(str(n)) == 1, stdout
        assert stderr == ""
        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


@flaky(max_runs=3)
def test_subprocess_noprint():
    """mp.Process without print doesn't trigger iostream mp_mode"""
    with kernel() as kc:

        np = 5
        code = "\n".join(
            [
                "import multiprocessing as mp",
                "pool = [mp.Process(target=range, args=(i,)) for i in range(%i)]" % np,
                "for p in pool: p.start()",
                "for p in pool: p.join()",
            ]
        )

        msg_id, content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == ""
        assert stderr == ""

        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


@flaky(max_runs=3)
@pytest.mark.skipif(
    sys.platform == "win32" or (sys.platform == "darwin" and sys.version_info >= (3, 8)),
    reason="subprocess prints fail on Windows and MacOS Python 3.8+",
)
def test_subprocess_error():
    """error in mp.Process doesn't crash"""
    with new_kernel() as kc:

        code = "\n".join(
            [
                "import multiprocessing as mp",
                "p = mp.Process(target=int, args=('hi',))",
                "p.start()",
                "p.join()",
            ]
        )

        msg_id, content = execute(kc=kc, code=code)
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == ""
        assert "ValueError" in stderr

        _check_master(kc, expected=True)
        _check_master(kc, expected=True, stream="stderr")


# raw_input tests


def test_raw_input():
    """test input"""
    with kernel() as kc:
        iopub = kc.iopub_channel

        input_f = "input"
        theprompt = "prompt> "
        code = f'print({input_f}("{theprompt}"))'
        msg_id = kc.execute(code, allow_stdin=True)
        msg = kc.get_stdin_msg(timeout=TIMEOUT)
        assert msg["header"]["msg_type"] == "input_request"
        content = msg["content"]
        assert content["prompt"] == theprompt
        text = "some text"
        kc.input(text)
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        stdout, stderr = assemble_output(kc.get_iopub_msg)
        assert stdout == text + "\n"


def test_save_history():
    # Saving history from the kernel with %hist -f was failing because of
    # unicode problems on Python 2.
    with kernel() as kc, TemporaryDirectory() as td:
        file = os.path.join(td, "hist.out")
        execute("a=1", kc=kc)
        wait_for_idle(kc)
        execute('b="abcþ"', kc=kc)
        wait_for_idle(kc)
        _, reply = execute("%hist -f " + file, kc=kc)
        assert reply["status"] == "ok"
        with open(file, encoding="utf-8") as f:
            content = f.read()
        assert "a=1" in content
        assert 'b="abcþ"' in content


def test_smoke_faulthandler():
    faulthadler = pytest.importorskip("faulthandler", reason="this test needs faulthandler")
    with kernel() as kc:
        # Note: faulthandler.register is not available on windows.
        code = "\n".join(
            [
                "import sys",
                "import faulthandler",
                "import signal",
                "faulthandler.enable()",
                'if not sys.platform.startswith("win32"):',
                "    faulthandler.register(signal.SIGTERM)",
            ]
        )
        _, reply = execute(code, kc=kc)
        assert reply["status"] == "ok", reply.get("traceback", "")


def test_help_output():
    """ipython kernel --help-all works"""
    cmd = [sys.executable, "-m", "IPython", "kernel", "--help-all"]
    proc = subprocess.run(cmd, timeout=30, capture_output=True)
    assert proc.returncode == 0, proc.stderr
    assert b"Traceback" not in proc.stderr
    assert b"Options" in proc.stdout
    assert b"Class" in proc.stdout


def test_is_complete():
    with kernel() as kc:
        # There are more test cases for this in core - here we just check
        # that the kernel exposes the interface correctly.
        kc.is_complete("2+2")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "complete"

        # SyntaxError
        kc.is_complete("raise = 2")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "invalid"

        kc.is_complete("a = [1,\n2,")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "incomplete"
        assert reply["content"]["indent"] == ""

        # Cell magic ends on two blank lines for console UIs
        kc.is_complete("%%timeit\na\n\n")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "complete"


@pytest.mark.skipif(sys.platform != "win32", reason="only run on Windows")
def test_complete():
    with kernel() as kc:
        execute("a = 1", kc=kc)
        wait_for_idle(kc)
        cell = "import IPython\nb = a."
        kc.complete(cell)
        reply = kc.get_shell_msg(timeout=TIMEOUT)

    c = reply["content"]
    assert c["status"] == "ok"
    start = cell.find("a.")
    end = start + 2
    assert c["cursor_end"] == cell.find("a.") + 2
    assert c["cursor_start"] <= end

    # there are many right answers for cursor_start,
    # so verify application of the completion
    # rather than the value of cursor_start

    matches = c["matches"]
    assert matches
    for m in matches:
        completed = cell[: c["cursor_start"]] + m
        assert completed.startswith(cell)


def test_matplotlib_inline_on_import():
    pytest.importorskip("matplotlib", reason="this test requires matplotlib")
    with kernel() as kc:
        cell = "\n".join(
            ["import matplotlib, matplotlib.pyplot as plt", "backend = matplotlib.get_backend()"]
        )
        _, reply = execute(cell, user_expressions={"backend": "backend"}, kc=kc)
        _check_status(reply)
        backend_bundle = reply["user_expressions"]["backend"]
        _check_status(backend_bundle)
        assert "backend_inline" in backend_bundle["data"]["text/plain"]


def test_message_order():
    N = 100  # number of messages to test
    with kernel() as kc:
        _, reply = execute("a = 1", kc=kc)
        _check_status(reply)
        offset = reply["execution_count"] + 1
        cell = "a += 1\na"
        msg_ids = []
        # submit N executions as fast as we can
        for _ in range(N):
            msg_ids.append(kc.execute(cell))
        # check message-handling order
        for i, msg_id in enumerate(msg_ids, offset):
            reply = kc.get_shell_msg(timeout=TIMEOUT)
            _check_status(reply["content"])
            assert reply["content"]["execution_count"] == i
            assert reply["parent_header"]["msg_id"] == msg_id


@pytest.mark.skipif(
    sys.platform.startswith("linux") or sys.platform.startswith("darwin"),
    reason="test only on windows",
)
def test_unc_paths():
    with kernel() as kc, TemporaryDirectory() as td:
        drive_file_path = os.path.join(td, "unc.txt")
        with open(drive_file_path, "w+") as f:
            f.write("# UNC test")
        unc_root = "\\\\localhost\\C$"
        file_path = os.path.splitdrive(os.path.dirname(drive_file_path))[1]
        unc_file_path = os.path.join(unc_root, file_path[1:])

        kc.execute(f"cd {unc_file_path:s}")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        out, err = assemble_output(kc.get_iopub_msg)
        assert unc_file_path in out

        flush_channels(kc)
        kc.execute(code="ls")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"
        out, err = assemble_output(kc.get_iopub_msg)
        assert "unc.txt" in out

        kc.execute(code="cd")
        reply = kc.get_shell_msg(timeout=TIMEOUT)
        assert reply["content"]["status"] == "ok"


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="does not work on PyPy",
)
def test_shutdown():
    """Kernel exits after polite shutdown_request"""
    with new_kernel() as kc:
        km = kc.parent
        execute("a = 1", kc=kc)
        wait_for_idle(kc)
        kc.shutdown()
        for _ in range(300):  # 30s timeout
            if km.is_alive():
                time.sleep(0.1)
            else:
                break
        assert not km.is_alive()


def test_interrupt_during_input():
    """
    The kernel exits after being interrupted while waiting in input().

    input() appears to have issues other functions don't, and it needs to be
    interruptible in order for pdb to be interruptible.
    """
    with new_kernel() as kc:
        km = kc.parent
        msg_id = kc.execute("input()")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)


@pytest.mark.skipif(os.name == "nt", reason="Message based interrupt not supported on Windows")
def test_interrupt_with_message():
    with new_kernel() as kc:
        km = kc.parent
        km.kernel_spec.interrupt_mode = "message"
        msg_id = kc.execute("input()")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)


@pytest.mark.skipif(
    "__pypy__" in sys.builtin_module_names,
    reason="fails on pypy",
)
def test_interrupt_during_pdb_set_trace():
    """
    The kernel exits after being interrupted while waiting in pdb.set_trace().

    Merely testing input() isn't enough, pdb has its own issues that need
    to be handled in addition.

    This test will fail with versions of IPython < 7.14.0.
    """
    with new_kernel() as kc:
        km = kc.parent
        msg_id = kc.execute("import pdb; pdb.set_trace()")
        msg_id2 = kc.execute("3 + 4")
        time.sleep(1)  # Make sure it's actually waiting for input.
        km.interrupt_kernel()
        from .test_message_spec import validate_message

        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id)
        # If we failed to interrupt interrupt, this will timeout:
        reply = get_reply(kc, msg_id2, TIMEOUT)
        validate_message(reply, "execute_reply", msg_id2)


def test_control_thread_priority():

    N = 5
    with new_kernel() as kc:
        msg_id = kc.execute("pass")
        get_reply(kc, msg_id)

        sleep_msg_id = kc.execute("import asyncio; await asyncio.sleep(2)")

        # submit N shell messages
        shell_msg_ids = []
        for i in range(N):
            shell_msg_ids.append(kc.execute(f"i = {i}"))

        # ensure all shell messages have arrived at the kernel before any control messages
        time.sleep(0.5)
        # at this point, shell messages should be waiting in msg_queue,
        # rather than zmq while the kernel is still in the middle of processing
        # the first execution

        # now send N control messages
        control_msg_ids = []
        for _ in range(N):
            msg = kc.session.msg("kernel_info_request", {})
            kc.control_channel.send(msg)
            control_msg_ids.append(msg["header"]["msg_id"])

        # finally, collect the replies on both channels for comparison
        get_reply(kc, sleep_msg_id)
        shell_replies = []
        for msg_id in shell_msg_ids:
            shell_replies.append(get_reply(kc, msg_id))

        control_replies = []
        for msg_id in control_msg_ids:
            control_replies.append(get_reply(kc, msg_id, channel="control"))

    # verify that all control messages were handled before all shell messages
    shell_dates = [msg["header"]["date"] for msg in shell_replies]
    control_dates = [msg["header"]["date"] for msg in control_replies]
    # comparing first to last ought to be enough, since queues preserve order
    # use <= in case of very-fast handling and/or low resolution timers
    assert control_dates[-1] <= shell_dates[0]


def _child():
    print("in child", os.getpid())

    def _print_and_exit(sig, frame):
        print(f"Received signal {sig}")
        # take some time so retries are triggered
        time.sleep(0.5)
        sys.exit(-sig)

    signal.signal(signal.SIGTERM, _print_and_exit)
    time.sleep(30)


def _start_children():
    ip = IPython.get_ipython()
    ns = ip.user_ns

    cmd = [sys.executable, "-c", f"from {__name__} import _child; _child()"]
    child_pg = Popen(cmd, start_new_session=False)
    child_newpg = Popen(cmd, start_new_session=True)
    ns["pid"] = os.getpid()
    ns["child_pg"] = child_pg.pid
    ns["child_newpg"] = child_newpg.pid
    # give them time to start up and register signal handlers
    time.sleep(1)


@pytest.mark.skipif(
    platform.python_implementation() == "PyPy",
    reason="does not work on PyPy",
)
@pytest.mark.skipif(
    sys.platform.lower() == "linux",
    reason="Stalls on linux",
)
def test_shutdown_subprocesses():
    """Kernel exits after polite shutdown_request"""
    with new_kernel() as kc:
        km = kc.parent
        msg_id, reply = execute(
            f"from {__name__} import _start_children\n_start_children()",
            kc=kc,
            user_expressions={
                "pid": "pid",
                "child_pg": "child_pg",
                "child_newpg": "child_newpg",
            },
        )
        print(reply)
        expressions = reply["user_expressions"]
        kernel_process = psutil.Process(int(expressions["pid"]["data"]["text/plain"]))
        child_pg = psutil.Process(int(expressions["child_pg"]["data"]["text/plain"]))
        child_newpg = psutil.Process(int(expressions["child_newpg"]["data"]["text/plain"]))
        wait_for_idle(kc)

        kc.shutdown()
        for _ in range(300):  # 30s timeout
            if km.is_alive():
                time.sleep(0.1)
            else:
                break
        assert not km.is_alive()
        assert not kernel_process.is_running()
        # child in the process group shut down
        assert not child_pg.is_running()
        # child outside the process group was not shut down (unix only)
        if os.name != "nt":
            assert child_newpg.is_running()
        try:
            child_newpg.terminate()
        except psutil.NoSuchProcess:
            pass



import os
from textwrap import dedent

import pytest
from flaky import flaky

from .test_embed_kernel import setup_kernel

TIMEOUT = 15

if os.name == "nt":
    pytest.skip("skipping tests on windows", allow_module_level=True)


@flaky(max_runs=3)
def test_ipython_start_kernel_userns():
    cmd = dedent(
        """
        from ipykernel.kernelapp import launch_new_instance
        ns = {"tre": 123}
        launch_new_instance(user_ns=ns)
        """
    )

    with setup_kernel(cmd) as client:
        client.inspect("tre")
        msg = client.get_shell_msg(timeout=TIMEOUT)
        content = msg["content"]
        assert content["found"]
        text = content["data"]["text/plain"]
        assert "123" in text

        # user_module should be an instance of DummyMod
        client.execute("usermod = get_ipython().user_module")
        msg = client.get_shell_msg(timeout=TIMEOUT)
        content = msg["content"]
        assert content["status"] == "ok"
        client.inspect("usermod")
        msg = client.get_shell_msg(timeout=TIMEOUT)
        content = msg["content"]
        assert content["found"]
        text = content["data"]["text/plain"]
        assert "DummyMod" in text


@flaky(max_runs=3)
def test_ipython_start_kernel_no_userns():
    # Issue #4188 - user_ns should be passed to shell as None, not {}
    cmd = dedent(
        """
        from ipykernel.kernelapp import launch_new_instance
        launch_new_instance()
        """
    )

    with setup_kernel(cmd) as client:
        # user_module should not be an instance of DummyMod
        client.execute("usermod = get_ipython().user_module")
        msg = client.get_shell_msg(timeout=TIMEOUT)
        content = msg["content"]
        assert content["status"] == "ok"
        client.inspect("usermod")
        msg = client.get_shell_msg(timeout=TIMEOUT)
        content = msg["content"]
        assert content["found"]
        text = content["data"]["text/plain"]
        assert "DummyMod" not in text



"""Test eventloop integration"""

import asyncio
import os
import sys
import threading
import time

import pytest
import tornado

from ipykernel.eventloops import enable_gui, loop_asyncio, loop_cocoa, loop_tk

from .utils import execute, flush_channels, start_new_kernel

KC = KM = None


def setup():
    """start the global kernel (if it isn't running) and return its client"""
    global KM, KC
    KM, KC = start_new_kernel()
    flush_channels(KC)


def teardown():
    assert KM is not None
    assert KC is not None
    KC.stop_channels()
    KM.shutdown_kernel(now=True)


async_code = """
from ipykernel.tests._asyncio_utils import async_func
async_func()
"""


@pytest.mark.skipif(tornado.version_info < (5,), reason="only relevant on tornado 5")
def test_asyncio_interrupt():
    assert KM is not None
    assert KC is not None
    flush_channels(KC)
    msg_id, content = execute("%gui asyncio", KC)
    assert content["status"] == "ok", content

    flush_channels(KC)
    msg_id, content = execute(async_code, KC)
    assert content["status"] == "ok", content

    KM.interrupt_kernel()

    flush_channels(KC)
    msg_id, content = execute(async_code, KC)
    assert content["status"] == "ok"


windows_skip = pytest.mark.skipif(os.name == "nt", reason="causing failures on windows")


@windows_skip
@pytest.mark.skipif(sys.platform == "darwin", reason="hangs on macos")
def test_tk_loop(kernel):
    def do_thing():
        time.sleep(1)
        try:
            kernel.app_wrapper.app.quit()
        # guard for tk failing to start (if there is no display)
        except AttributeError:
            pass

    t = threading.Thread(target=do_thing)
    t.start()
    # guard for tk failing to start (if there is no display)
    try:
        loop_tk(kernel)
    except Exception:
        pass
    t.join()


@windows_skip
def test_asyncio_loop(kernel):
    def do_thing():
        loop.call_soon(loop.stop)

    loop = asyncio.get_event_loop()
    loop.call_soon(do_thing)
    loop_asyncio(kernel)


@windows_skip
def test_enable_gui(kernel):
    enable_gui("tk", kernel)


@pytest.mark.skipif(sys.platform != "darwin", reason="MacOS-only")
def test_cocoa_loop(kernel):
    loop_cocoa(kernel)



from ipykernel.comm import Comm, CommManager
from ipykernel.ipkernel import IPythonKernel


def test_comm(kernel):
    manager = CommManager(kernel=kernel)
    kernel.comm_manager = manager

    c = Comm(kernel=kernel, target_name="bar")
    msgs = []

    def on_close(msg):
        msgs.append(msg)

    def on_message(msg):
        msgs.append(msg)

    c.publish_msg("foo")
    c.open({})
    c.on_msg(on_message)
    c.on_close(on_close)
    c.handle_msg({})
    c.handle_close({})
    c.close()
    assert len(msgs) == 2
    assert c.target_name == "bar"


def test_comm_manager(kernel):
    manager = CommManager(kernel=kernel)
    msgs = []

    def foo(comm, msg):
        msgs.append(msg)
        comm.close()

    def fizz(comm, msg):
        raise RuntimeError('hi')

    def on_close(msg):
        msgs.append(msg)

    def on_msg(msg):
        msgs.append(msg)

    manager.register_target("foo", foo)
    manager.register_target("fizz", fizz)

    kernel.comm_manager = manager
    comm = Comm()
    comm.on_msg(on_msg)
    comm.on_close(on_close)
    manager.register_comm(comm)

    assert manager.get_comm(comm.comm_id) == comm
    assert manager.get_comm('foo') is None

    msg = dict(content=dict(comm_id=comm.comm_id, target_name='foo'))
    manager.comm_open(None, None, msg)
    assert len(msgs) == 1
    msg['content']['target_name'] = 'bar'
    manager.comm_open(None, None, msg)
    assert len(msgs) == 1
    msg = dict(content=dict(comm_id=comm.comm_id, target_name='fizz'))
    manager.comm_open(None, None, msg)
    assert len(msgs) == 1

    manager.register_comm(comm)
    assert manager.get_comm(comm.comm_id) == comm
    msg = dict(content=dict(comm_id=comm.comm_id))
    manager.comm_msg(None, None, msg)
    assert len(msgs) == 2
    msg['content']['comm_id'] = 'foo'
    manager.comm_msg(None, None, msg)
    assert len(msgs) == 2

    manager.register_comm(comm)
    assert manager.get_comm(comm.comm_id) == comm
    msg = dict(content=dict(comm_id=comm.comm_id))
    manager.comm_close(None, None, msg)
    assert len(msgs) == 3

    assert comm._closed


def test_comm_in_manager(ipkernel: IPythonKernel) -> None:
    comm = Comm()
    assert comm.comm_id in ipkernel.comm_manager.comms
